@@ -147,6 +147,7 @@ def insert_optional_args(args):
 		'--jobs'       : valid_integers,
 		'--keep-going'           : y_or_n,
 		'--load-average'         : valid_floats,
+		'--onlydeps-with-rdeps'  : y_or_n,
 		'--package-moves'        : y_or_n,
 		'--quiet'                : y_or_n,
 		'--quiet-build'          : y_or_n,
@@ -326,6 +327,12 @@ def parse_opts(tmpcmdline, silent=False):
 			"choices" : true_y_or_n
 		},
 
+		"--autounmask-backtrack": {
+			"help": ("continue backtracking when there are autounmask "
+				"configuration changes"),
+			"choices":("y", "n")
+		},
+
 		"--autounmask-continue": {
 			"help"    : "write autounmask changes and continue",
 			"choices" : true_y_or_n
@@ -559,6 +566,11 @@ def parse_opts(tmpcmdline, silent=False):
 			"action" : "append",
 		},
 
+		"--onlydeps-with-rdeps": {
+			"help"    : "modify interpretation of depedencies",
+			"choices" : true_y_or_n
+		},
+
 		"--rebuild-exclude": {
 			"help"   :"A space separated list of package names or slot atoms. " + \
 				"Emerge will not rebuild these packages due to the " + \
